/* the gmp3 project.
 * (c)2002 by Samuel Abels (sam@manicsadness.com)
 * This project's homepage is: http://software.manicsadness.com/gmp3
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <gnome.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <ctype.h>
#include "mp3info.h"
#include "lib_id3v2.3.h"

/*
 * Purpose: Reads the ID3 tag from a file.
 * Parameters: tag - The structure to store the tag in, filename - The name of the file to operate on.
 * Returns:
 * 0 if successful,
 * 1 if an error occured when opening the file
 * 2 if error while reading tag.
 * 3 if no TAG found.
 */
int get_id3v2_tag(id3Tag *tag, gchar *filename)
{
	id3v2Tag *v2 = calloc(1, sizeof(id3v2Tag));
	id3v2Frame *frame = NULL;
	FILE *mp3file;
	guchar *c = malloc(sizeof(guchar));
	gint i = -1;
	GList *frames = NULL;
	GList *curframe = NULL;

	memset(v2, 0, sizeof(id3v2Tag));

// open file
	mp3file = fopen(filename, "rb");
	if (!mp3file)
	{
		free(c);
		free(v2);
		return (1);
	}

// get header (=first ten bytes)
	fseek(mp3file, 0, SEEK_SET);
	if( fread(c, 1, 10, mp3file) < 10 )
	{
		free(c);
		free(v2);
		fclose(mp3file);
		return (1);
	}

// is a tag present?
	if(strncmp(c, "ID3", 3))
	{
		free(c);
		free(v2);
		fclose(mp3file);
		return (2);
	}

	v2->unsync = (c[5] & 128) >> 7;
	v2->has_extheader = (c[5] & 64) >> 6;
	v2->is_experimental = (c[5] & 32) >> 5;
	v2->has_footer = (c[5] & 32) >> 5;
	
 	v2->tag_size = c[9] + (c[8] << 7) + (c[7] << 14) + (c[6] << 21) + 10;
	if(v2->has_footer) v2->tag_size += 10;
	
// check if version is supported
	if(c[3] != 3 && c[3] != 4)
	{
		free(c);
		free(v2);
		fclose(mp3file);
		return (3);
	}
// get extheader
	if(v2->has_extheader)
	{
		if( fread(c, 1, 5, mp3file) < 5 )
		{
			free(c);
			free(v2);
			fclose(mp3file);
			return (4);
		}
		v2->extheader_size = c[3] + (c[2] << 7) + (c[1] << 14) + (c[0] << 21);
		v2->flag_bytes = (c[4] > 0) ? c[4] : 1;
		if( fread(c, 1, v2->flag_bytes, mp3file) < v2->flag_bytes )
		{
			free(c);
			free(v2);
			fclose(mp3file);
			return (5);
		}
		v2->is_update = (c[0] & 64) >> 6; 
		v2->crc_data_present = (c[0] & 32) >> 5;
		v2->restrictions = (c[0] & 16) >> 4;
		
		if(v2->is_update) fread(c, 1, 1, mp3file);
// crc data flag byte
		if(v2->crc_data_present)
		{
			fread(c, 1, 1, mp3file);
			if(*c != 5)
			{
				free(c);
				free(v2);
				fclose(mp3file);
				return (6);
			}
			v2->crc_data_length = *c;
			v2->crc_data = calloc(*c, sizeof(guchar));
			fread(v2->crc_data, *c, sizeof(guchar), mp3file);
		}
// restrictions flag byte
		if(v2->restrictions)
		{
			fread(c, 1, 1, mp3file);
			if(*c != 1)
			{
				free(c);
				if(v2->crc_data)
					free(v2->crc_data);
				free(v2);
				fclose(mp3file);
				return (7);
			}
			v2->restrictions_data_length = *c;
			v2->restrictions_data = calloc(*c, sizeof(guchar));
			fread(v2->restrictions_data, *c, sizeof(guchar), mp3file);
		}
	}

// get the content frames
	i = -1;
	while(ftell(mp3file) < v2->tag_size)
	{
		fread(c, 10, sizeof(guchar), mp3file);

		if(c[0] == 0 && c[1] == 0 && c[2] ==  0 && c[3] == 0)
			break;
		if(!isalnum(c[0]) || !isalnum(c[1]) || !isalnum(c[2]) || !isalnum(c[3]))
			break;
		
		frame = malloc(sizeof(id3v2Frame));

		frame->frame_id = calloc(5, 5);
		strncpy(frame->frame_id, c, 4);
		frame->frame_datasize = c[7] + (c[6] << 7) + (c[5] << 14) + (c[4] << 21);
		frame->frame_statusflag = c[8];
		frame->frame_formatflag = c[9];
		
		frames = g_list_append(frames, frame);
		
		if(frame->frame_datasize > 60000)
			break;
		
		frame->frame_data = calloc(frame->frame_datasize + 1, sizeof(guchar));
		fread(frame->frame_data, frame->frame_datasize, sizeof(guchar), mp3file);
		
	}
	v2->frames = frames;
	
	free(c);
	fclose(mp3file);

// at last we copy the raw data to the users struct
	tag->size = v2->tag_size;
	memset(tag->artist, 0, 1024);
	memset(tag->title, 0, 1024);
	memset(tag->album, 0, 1024);
	memset(tag->year, 0, 5);
	memset(tag->comment, 0, 1024);
	memset(tag->track, 0, 3);
	tag->has_footer = v2->has_footer;
	
	curframe = g_list_first(frames);
	while(curframe)
	{
		GList *tempframe = NULL;
		frame = (id3v2Frame *)curframe->data;
		
		if(strncmp(frame->frame_id, "TPE1", 4)==0)
			strncpy(tag->artist, frame->frame_data + 1, 1023);
		if(strncmp(frame->frame_id, "TIT2", 4)==0)
			strncpy(tag->title, frame->frame_data + 1, 1023);
		if(strncmp(frame->frame_id, "TALB", 4)==0)
			strncpy(tag->album, frame->frame_data + 1, 1023);
		if(strncmp(frame->frame_id, "TYER", 4)==0)
			strncpy(tag->year, frame->frame_data + 1, 4);
		if(strncmp(frame->frame_id, "COMM", 4)==0)
			strncpy(tag->comment, frame->frame_data + 1, 1023);
		if(strncmp(frame->frame_id, "TRCK", 4)==0)
			strncpy(tag->track, frame->frame_data + 1, 3);
		free(frame->frame_id);
		free(frame->frame_data);
		free(frame);
		
		tempframe = curframe;
		curframe = curframe->next;
		g_list_free_1(tempframe);
	}
	if(v2->crc_data)
		free(v2->crc_data);
	if(v2->restrictions_data)
		free(v2->restrictions_data);
	free(v2);
	return (0);
}




int
del_id3v2_tag(gchar *filename)
{
	id3Tag *tag = calloc(1, sizeof(id3Tag));
	long file_len;
	FILE *file;
	gint read;
	void *ptr;

// check if an valid old id3v2 tag is present
	if( !tag ||
		(get_id3v2_tag(tag, filename) != 0
		&& get_id3v2_tag(tag, filename) != 3) ) return 0;
	
	ptr =  malloc(4096);

// open file read/write
	file = fopen(filename, "r+b");
	if (!file) return(1);

	fseek(file, 0, SEEK_END);
	file_len = ftell(file);
	if(file_len < 1 || tag->size < 1) return 1;

// set anything but the header in tag to zero. ill not really remove the tag,
// because this would be much slower and if we write a new tag, this would mean wed have to
// rewrite the complete tag.
	fseek(file, 10, SEEK_SET);
	for(read=0;read<tag->size-10;read++)
	{
		fputc(0, file);
	}

	fclose(file);
	free(ptr);
	free(tag);
	
	return 0;
}

















gint
set_id3v2_tag(id3Tag *tag, gchar *filename)
{
	guchar *btag, *btag_start;
	guchar flag = 0;
	gint i, fd;
	gchar *d;
	id3Tag *old = calloc(1,sizeof(id3Tag));
	id3v2Tag *v2 = calloc(1,sizeof(id3v2Tag));
	id3v2Frame *frame = NULL;
	GList *frames = NULL;
	GList *curframe = NULL;

// first of all ill generate a valid id3v2 tag struct out of the tag struct we got by the user.
// Set the flags...
	v2->unsync = FALSE;
	v2->is_experimental = FALSE;
	v2->has_footer = FALSE;
	v2->tag_size = strlen(tag->title) + 10
		+ strlen(tag->artist) + 10
		+ strlen(tag->album) + 10
		+ strlen(tag->year) + 10
		+ strlen(tag->comment) + 10
		+ strlen(tag->track) + 10
		+ 30; // Genre
	v2->is_update = FALSE;
	v2->crc_data_present = FALSE;
	v2->crc_data_length = 0;
	v2->restrictions = FALSE;
	v2->restrictions_data_length = 0;
// Set the contentframes
	i = -1;
	if(strlen(tag->title)>0)
	{
		++i;
		frame = malloc(sizeof(id3v2Frame));
		frame->frame_id = malloc(4);
		strncpy(frame->frame_id, "TIT2", 4);
		frame->frame_datasize = strlen(tag->title) + 1;
		frame->frame_statusflag = 0;
		frame->frame_formatflag = 0;
		frame->frame_data = malloc(strlen(tag->title) + 1);
		strncpy(frame->frame_data, tag->title, strlen(tag->title));
		frames = g_list_append(frames, frame);
	}
	
	if(strlen(tag->artist)>0)
	{
		++i;
		frame = malloc(sizeof(id3v2Frame));
		frame->frame_id = malloc(4);
		strncpy(frame->frame_id, "TPE1", 4);
		frame->frame_datasize = strlen(tag->artist) + 1;
		frame->frame_statusflag = 0;
		frame->frame_formatflag = 0;
		frame->frame_data = malloc(strlen(tag->artist) + 1);
		strncpy(frame->frame_data, tag->artist, strlen(tag->artist));
		frames = g_list_append(frames, frame);
	}
	
	if(strlen(tag->album)>0)
	{
		++i;
		frame = malloc(sizeof(id3v2Frame));
		frame->frame_id = malloc(4);
		strncpy(frame->frame_id, "TALB", 4);
		frame->frame_datasize = strlen(tag->album) + 1;
		frame->frame_statusflag = 0;
		frame->frame_formatflag = 0;
		frame->frame_data = malloc(strlen(tag->album) + 1);
		strncpy(frame->frame_data, tag->album, strlen(tag->album));
		frames = g_list_append(frames, frame);
	}
	
	if(strlen(tag->year)>0)
	{
		++i;
		frame = malloc(sizeof(id3v2Frame));
		frame->frame_id = malloc(4);
		strncpy(frame->frame_id, "TYER", 4);
		frame->frame_datasize = strlen(tag->year) + 1;
		frame->frame_statusflag = 0;
		frame->frame_formatflag = 0;
		frame->frame_data = malloc(strlen(tag->year) + 1);
		strncpy(frame->frame_data, tag->year, strlen(tag->year));
		frames = g_list_append(frames, frame);
	}
	
	if(strlen(tag->comment)>0)
	{
		++i;
		frame = malloc(sizeof(id3v2Frame));
		frame->frame_id = malloc(4);
		strncpy(frame->frame_id, "COMM", 4);
		frame->frame_datasize = strlen(tag->comment) + 1;
		frame->frame_statusflag = 0;
		frame->frame_formatflag = 0;
		frame->frame_data = malloc(strlen(tag->comment) + 1);
		strncpy(frame->frame_data, tag->comment, strlen(tag->comment));
		frames = g_list_append(frames, frame);
	}
	
	if(atoi(tag->track) > 0)
	{
		++i;
		frame = malloc(sizeof(id3v2Frame));
		frame->frame_id = malloc(4);
		strncpy(frame->frame_id, "TRCK", 4);
		frame->frame_datasize = 4;
		frame->frame_statusflag = 0;
		frame->frame_formatflag = 0;
		frame->frame_data = malloc(4);
		if(atoi(tag->track) < 10)
			snprintf(frame->frame_data, 3, "0%i", atoi(tag->track));
		else
			snprintf(frame->frame_data, 3, "%i", atoi(tag->track));
		frames = g_list_append(frames, frame);
	}
	
	v2->flag_bytes = i + 1;

// check if an old v2 is present and get its length
	if(get_id3v2_tag(old, filename)!=0)
	{
		free(old);
		old = NULL;
	}
	else
	{
		if( ((gint)old->size - (gint)v2->tag_size) < 0 )
		{
			free(old);
			old = NULL;
		}
	}		

	fd = open(filename, O_RDWR);
	if(fd < 0)
	{
		curframe = g_list_first(frames);
		while(curframe)
		{
			frame = (id3v2Frame *)curframe->data;
			
			free(frame->frame_id);
			free(frame->frame_data);
			free(frame);
			
			curframe = curframe->next;
		}
		g_list_free(frames);
		free(old);
		free(v2);
		return 1;
	}

// now we are going to write the tags raw data into the btag byte array
	btag = btag_start = malloc(v2->tag_size);
	memset(btag, 0, v2->tag_size);
	strncpy(btag, "ID3", 3);
	btag += 3;
	*btag++ = 3;	// major version
	*btag++ = 0;	// minor version
	flag |= ((v2->unsync & 1) << 7);
	flag |= ((v2->has_extheader & 1) << 6);
	flag |= ((v2->is_experimental & 1) << 5);
	flag |= ((v2->has_footer & 1) << 4);
	memcpy(btag++, &flag, 1);;
	
	if(old)
	{
		i = old->size - 10;
		if(old->has_footer) i -= 10;
	}
	else
	{
		i = v2->tag_size - 10;
		if(v2->has_footer) i -= 10;
// add a padding to total size we mean to store on disk
		i += 1024;
	}
	
// the tag size must be syncsafe
	d = malloc(4);
	d[3] = (i & 0x7f);
	d[2] = ((i & 0x80) >> 7) | (((i & 0x7f00) >> 8) << 1);
	d[1] = ((i & 0x8000) >> 15) | (((i & 0x7f0000) >> 16) << 1);
	d[0] = ((i & 0x800000) >> 23) | (((i & 0x7f000000) >> 24) << 1);
	btag[0] = d[0];
	btag[1] = d[1];
	btag[2] = d[2];
	btag[3] = d[3];
	free(d);
	btag += 4;

// and last of all, generate the contentflags (and syncsafe and blabla....)
	curframe = g_list_first(frames);
	while(curframe)
	{
		frame = (id3v2Frame *)curframe->data;
		
		strncpy(btag, frame->frame_id, 4);
		btag += 4;
		d = malloc(4);
		d[3] = (frame->frame_datasize & 0x7f);
		d[2] = ((frame->frame_datasize & 0x80) >> 7) | (((frame->frame_datasize & 0x7f00) >> 8) << 1);
		d[1] = ((frame->frame_datasize & 0x8000) >> 15) | (((frame->frame_datasize & 0x7f0000) >> 16) << 1);
		d[0] = ((frame->frame_datasize & 0x800000) >> 23) | (((frame->frame_datasize & 0x7f000000) >> 24) << 1);
		btag[0] = d[0];
		btag[1] = d[1];
		btag[2] = d[2];
		btag[3] = d[3];
		free(d);
		btag += 4;
		memcpy(btag++, &frame->frame_statusflag, 1);
		memcpy(btag++, &frame->frame_formatflag, 1);
		memcpy(btag++, "\0", 1);
		
		memcpy(btag, frame->frame_data, frame->frame_datasize - 1);
		btag += frame->frame_datasize - 1;
		
		curframe = curframe->next;
	}
	
// if an old v2 was provided, overwrite it
	if(old)
	{  
		FILE *file = NULL;
		gpointer ptr = calloc(old->size - v2->tag_size, old->size - v2->tag_size);
		if(!(file = fdopen(fd, "r+b")))
		{
			curframe = g_list_first(frames);
			while(curframe)
			{
				frame = (id3v2Frame *)curframe->data;
				
				free(frame->frame_id);
				free(frame->frame_data);
				free(frame);
				
				curframe = curframe->next;
			}
			g_list_free(frames);
			free(btag_start);
			free(old);
			free(v2);
			return 11;
		}
		
		fseek(file, 0, SEEK_SET);
		if(fwrite(btag_start, v2->tag_size, 1, file) < 1)
		{
			curframe = g_list_first(frames);
			while(curframe)
			{
				frame = (id3v2Frame *)curframe->data;
				
				free(frame->frame_id);
				free(frame->frame_data);
				free(frame);
				
				curframe = curframe->next;
			}
			g_list_free(frames);
			free(btag_start);
			free(old);
			free(v2);
			return 12;
		}
			
// write padding till end of old v2
		if(fwrite(ptr, old->size - v2->tag_size, 1, file) < 1)
		{
			curframe = g_list_first(frames);
			while(curframe)
			{
				frame = (id3v2Frame *)curframe->data;
				
				free(frame->frame_id);
				free(frame->frame_data);
				free(frame);
				
				curframe = curframe->next;
			}
			g_list_free(frames);
			free(btag_start);
			free(old);
			free(v2);
			return 13;
		}
		
		fflush(file);
		free(ptr);
		
	} else {
// otherwise create a completely new one, which will mean to completely rewrite the file
		FILE *file, *tmp;
		int read;
		void *ptr, *blank;

		ptr = malloc(4096);
		blank = calloc(1024, sizeof(gchar));

		file = fdopen(fd, "r+b");
		tmp = tmpfile();
		if(!(file && tmp))
		{
			curframe = g_list_first(frames);
			while(curframe)
			{
				frame = (id3v2Frame *)curframe->data;
				
				free(frame->frame_id);
				free(frame->frame_data);
				free(frame);
				
				curframe = curframe->next;
			}
			g_list_free(frames);
			free(btag_start);
			free(old);
			free(v2);
			return 14;
		}

		fseek(file, 0, SEEK_SET);
		fseek(tmp, 0, SEEK_SET);

// write tag to tempfile
		fwrite(btag_start, v2->tag_size, 1, tmp);

// Write 1024b padding
		fwrite(blank, 1024, 1, tmp);

// write rest of file
		while(!feof(file))
		{
			read = fread(ptr, 1, 4096, file);
			if(fwrite(ptr, 1, read, tmp) != read && !feof(file))
			{
				curframe = g_list_first(frames);
				while(curframe)
				{
					frame = (id3v2Frame *)curframe->data;
					
					free(frame->frame_id);
					free(frame->frame_data);
					free(frame);
					
					curframe = curframe->next;
				}
				g_list_free(frames);
				free(btag_start);
				free(old);
				free(v2);
				return 15;
			}
		}

		fflush(tmp);

		fseek(file, 0, SEEK_SET);
		fseek(tmp, 0, SEEK_SET);
		while(!feof(tmp))
		{
			read = fread(ptr, 1, 4096, tmp);
			if(fwrite(ptr, 1, read, file) != read && !feof(tmp))
			{
				curframe = g_list_first(frames);
				while(curframe)
				{
					frame = (id3v2Frame *)curframe->data;
					
					free(frame->frame_id);
					free(frame->frame_data);
					free(frame);
					
					curframe = curframe->next;
				}
				g_list_free(frames);
				free(btag_start);
				free(old);
				free(v2);
				return 16;
			}
		}

		fflush(file);
		fclose(tmp);

		free(ptr);
		free(blank);
	}

	curframe = g_list_first(frames);
	while(curframe)
	{
		frame = (id3v2Frame *)curframe->data;
		
		free(frame->frame_id);
		free(frame->frame_data);
		free(frame);
		
		curframe = curframe->next;
	}
	g_list_free(frames);

	free(btag_start);
	free(old);
	free(v2);
	return 0;
}
